#include <sta/config.hpp>
#ifdef STA_CAN_BUS_ENABLE

#include <sta/tacos/can_bus.hpp>
#include <sta/debug/assert.hpp>
#include <sta/tacos/manager.hpp>

// TODO fix this shit
extern CAN_HandleTypeDef hcan1;
namespace sta
{
	namespace tacos
	{
		CanBus::CanBus()
			: TacosThread{"Can Bus", STA_TACOS_CAN_BUS_PRIORITY},
			  canBusController_(new STM32CanController(&STA_STM32_CAN_HANDLE)),
			  canBusSysQueue_(STA_RTOS_CAN_BUS_QUEUE_LENGTH),
			  canBusDataQueue_(STA_RTOS_CAN_BUS_QUEUE_LENGTH),
              canBus_{canBusController_, HAL_GetTick, sta::tacos::handleSysMessage, sta::tacos::handleDataMessage}
		{
			bufferIndex = 0;
			for(int i = 0; i < STA_RTOS_CAN_BUS_QUEUE_LENGTH; i++){
				canBusSysQueueBuffer_[i] = nullptr;
			}
		}

        void CanBus::init()
	    {
		    canBusController_->start();

		    if (HAL_CAN_ActivateNotification(&hcan1, CAN_IT_RX_FIFO0_MSG_PENDING) != HAL_OK ||
		    		HAL_CAN_ActivateNotification(&hcan1, CAN_IT_RX_FIFO1_MSG_PENDING))
		    {
		      	 Error_Handler();
		    }
	    }

        void CanBus::func()
        {
        	messageEvent.clear(STA_RTOS_CAN_ANY);
            uint32_t flags = messageEvent.wait(STA_RTOS_CAN_ANY, osWaitForever);

            if (flags != static_cast<uint32_t>(osErrorTimeout))
            {
                STA_ASSERT_MSG((flags & osStatusReserved) == flags, "Unexpected error occurred in wait");

                if (flags & STA_RTOS_CAN_FLAG_SYS_QUEUED)
                {
                    // Take messages from queue until empty
                    CanSysMsg msg;
                    while (CanBus::_instance->getCanBusMsg(&msg, 0))
                    {
                        canBus_.send(msg);
                    }
                }

                if (flags & STA_RTOS_CAN_FLAG_MSG_AVAIL)
                {
                    CanSysMsg sysMsg;
                    // Iterate through buffer and set back to nullptr after use
                    for(int i = 0; i < STA_RTOS_CAN_BUS_QUEUE_LENGTH; i++){
                    	if(canBusSysQueueBuffer_[i] != nullptr){
                    		sysMsg = *canBusSysQueueBuffer_[i];
                    		canBusSysQueueBuffer_[i] = nullptr;

                    		// Append to the correct thread's queue
							for (std::shared_ptr<TacosThread> thread : Manager::instance()->getActiveThreads()){
								if (thread->getCanID() == sysMsg.header.sid){
									thread->CAN_queue_.put(sysMsg);
									break;
								}
							}
						}
                    }
                    //canBus_.processRx();
                }

                if (flags & STA_RTOS_CAN_FLAG_SHOW_STATS)
                {
                    canBus_.showStatistics();
                }
            }

        }

        bool CanBus::queueCanBusMsg(const CanDataMsg& msg, uint32_t timeout)
		{
			STA_ASSERT((msg.header.sid & STA_CAN_SID_SYS_BITS) == 0);
			STA_ASSERT(msg.header.payloadLength <= sizeof(msg.payload));

			if (canBusDataQueue_.put(msg, timeout))
			{
				// Signal task
				messageEvent.set(STA_RTOS_CAN_FLAG_DATA_QUEUED);
				messageEvent.clear(STA_RTOS_CAN_ANY);
				return true;
			}
			else
			{
				return false;
			}
		}

		bool CanBus::queueCanBusMsg(const CanSysMsg& msg, uint32_t timeout)
		{
			// This technically should check if we are using a system message, but we just pretending that everything is one of those rn
			//STA_ASSERT((msg.header.sid & ~STA_CAN_SID_SYS_BITS) == 0);

			if (canBusSysQueue_.put(msg, timeout))
			{
				// Signal task
				messageEvent.set(STA_RTOS_CAN_FLAG_SYS_QUEUED);
				messageEvent.clear(STA_RTOS_CAN_ANY);
				return true;
			}
			else
			{
				return false;
			}
		}

		void CanBus::canCallback(uint32_t fifo){
			if(messageEvent.get() != STA_RTOS_CAN_FLAG_MSG_AVAIL){
				// get here does not work since FreeRTOS is a buggy mess
				messageEvent.set(STA_RTOS_CAN_FLAG_MSG_AVAIL);

				CanRxHeader rxHeader;					//CAN Bus Receive Header
				uint8_t canRX[8] = {0,0,0,0,0,0,0,0};  	//CAN Bus Receive Buffer

				bool received_ = canBusController_->receiveFrame(fifo, &rxHeader, canRX);

				if(received_){
					CanSysMsg sysMsg;
					sysMsg.header.sid = rxHeader.id.sid;
					sysMsg.header.payloadLength = rxHeader.payloadLength;
					std::memcpy(sysMsg.payload, canRX, rxHeader.payloadLength);

					canBusSysQueueBuffer_[bufferIndex] = &sysMsg;

					bufferIndex++;
					if (bufferIndex >= STA_RTOS_CAN_BUS_QUEUE_LENGTH) bufferIndex = 0;
				}
			}
		}

		bool CanBus::getCanBusMsg(CanDataMsg * msg, uint32_t timeout)
		{
			return (canBusDataQueue_.get(msg, timeout));
		}

		bool CanBus::getCanBusMsg(CanSysMsg * msg, uint32_t timeout)
		{
			return (canBusSysQueue_.get(msg, timeout));
		}

		CanBus* CanBus::_instance = nullptr;

		RtosEvent CanBus::messageEvent;

    } /* namespace tacos */
} /* namespace sta */


namespace sta {

	void CanBus_RxPendingCallback(uint32_t fifo){
			sta::tacos::CanBus::instance()->canCallback(fifo);
	}

	namespace tacos
	{
		void handleSysMessage(const sta::CanRxHeader & header, const uint8_t * payload)
		{
			// Write frame payload to DebugSerial
			STA_DEBUG_PRINTLN("[event] RX sys frame");

			debug::printFrameID(header.id);
			debug::printPayloadHex(payload, header.payloadLength);

			// Sysmessage is mainly only state change from GRSM
			// TODO add other cases

			// 0 is from state, 1 is to state, 2 is lockout, 3 is failsafe (-1 if inactive)
			if(payload[1] > 0 && payload[1] < STA_TACOS_NUM_STATES){

				if(payload[3] == -1){
					sta::tacos::Statemachine::instance()->requestStateTransition(payload[0], payload[1], payload[2]);
				}
				else{
					sta::tacos::Statemachine::instance()->requestTimedStateTransition(payload[0], payload[1], payload[2], payload[3]);
				}
			}
		}

		void handleDataMessage(const sta::IsotpMessage & msg)
		{
			STA_ASSERT(msg.buffer != nullptr);
			STA_ASSERT(msg.size != 0);

			STA_DEBUG_PRINTLN("[event] RX data message");

			debug::printFrameID(msg.frameID);

			// TODO Forward message to other threads

	//		if (buffer[0] == DEMO_BMP_PACKET_ID)
	//		{
	//			BmpPacket packet;
	//			if (unpack(buffer + 1, size - 1, &packet))
	//			{
	//				STA_DEBUG_PRINTLN();
	//				STA_DEBUG_PRINTLN("# ############");
	//				STA_DEBUG_PRINTLN("# ## BMP380 ##");
	//				STA_DEBUG_PRINTLN("# ############");
	//				STA_DEBUG_PRINTLN("#");
	//
	//				STA_DEBUG_PRINT("# temperature: ");
	//				STA_DEBUG_PRINT(packet.temperature);
	//				STA_DEBUG_PRINTLN(" *C");
	//				STA_DEBUG_PRINT("# pressure: ");
	//				STA_DEBUG_PRINT(packet.pressure);
	//				STA_DEBUG_PRINTLN(" Pa");
	//				STA_DEBUG_PRINT("# altitude: ");
	//				STA_DEBUG_PRINT(packet.altitude);
	//				STA_DEBUG_PRINTLN(" m");
	//				STA_DEBUG_PRINTLN();
	//			}
	//			else
	//			{
	//				STA_DEBUG_PRINTLN("[error] BMP unpack failed");
	//			}
	//		}
	//		else
			{
				STA_DEBUG_PRINT("ID: ");
				STA_DEBUG_PRINTLN(msg.buffer[0], sta::IntegerBase::HEX);
				STA_DEBUG_PRINT("size: ");
				STA_DEBUG_PRINTLN(msg.size);
			}
		}
	} // namespace tacos
} // namespace sta

#endif // STA_CAN_BUS_ENABLE
